@@ -1,19 +1,16 @@
-# -*- coding: utf-8 -*-
-
-## This file is part of Gajim.
-##
-## Gajim is free software; you can redistribute it and/or modify
-## it under the terms of the GNU General Public License as published
-## by the Free Software Foundation; version 3 only.
-##
-## Gajim is distributed in the hope that it will be useful,
-## but WITHOUT ANY WARRANTY; without even the implied warranty of
-## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-## GNU General Public License for more details.
-##
-## You should have received a copy of the GNU General Public License
-## along with Gajim.  If not, see <http://www.gnu.org/licenses/>.
-##
+# This file is part of Gajim.
+#
+# Gajim is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published
+# by the Free Software Foundation; version 3 only.
+#
+# Gajim is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with Gajim.  If not, see <http://www.gnu.org/licenses/>.
 
 '''
 Plug-in management related classes.
@@ -30,16 +27,22 @@ import os
 import sys
 import fnmatch
 import zipfile
-from shutil import rmtree
+from shutil import rmtree, move
 import configparser
-from pkg_resources import parse_version
+from packaging.version import Version as V
 
+import gajim
 from gajim.common import app
 from gajim.common import nec
 from gajim.common import configpaths
+from gajim.common import modules
+from gajim.common.nec import NetworkEvent
+from gajim.common.i18n import _
 from gajim.common.exceptions import PluginsystemError
+from gajim.common.helpers import Singleton
+from gajim.plugins import plugins_i18n
 
-from gajim.plugins.helpers import log, log_calls, Singleton
+from gajim.plugins.helpers import log
 from gajim.plugins.helpers import GajimPluginActivateException
 from gajim.plugins.gajimplugin import GajimPlugin, GajimPluginException
 
@@ -59,7 +62,7 @@ class PluginManager(metaclass=Singleton):
     :todo: implement mechanism to dynamically load plugins where GUI extension
                points have been already called (i.e. when plugin is activated
                after GUI object creation). [DONE?]
-    :todo: implement mechanism to dynamically deactive plugins (call plugin's
+    :todo: implement mechanism to dynamically deactivate plugins (call plugin's
                deactivation handler) [DONE?]
     :todo: when plug-in is deactivated all GUI extension points are removed
                from `PluginManager.gui_extension_points_handlers`. But when
@@ -72,9 +75,6 @@ class PluginManager(metaclass=Singleton):
                'destructors' (classes that register GUI extension points)
     '''
 
-    __metaclass__ = Singleton
-
-    #@log_calls('PluginManager')
     def __init__(self):
         self.plugins = []
         '''
@@ -108,35 +108,78 @@ class PluginManager(metaclass=Singleton):
         Registered names with instances of encryption Plugins.
         '''
 
-        for path in [app.PLUGINS_DIRS[1], app.PLUGINS_DIRS[0]]:
-            pc = PluginManager.scan_dir_for_plugins(path)
+        self.update_plugins()
+
+        for path in reversed(configpaths.get_plugin_dirs()):
+            pc = self.scan_dir_for_plugins(path)
             self.add_plugins(pc)
 
-    @log_calls('PluginManager')
     def _plugin_has_entry_in_global_config(self, plugin):
         if app.config.get_per('plugins', plugin.short_name) is None:
             return False
-        else:
-            return True
+        return True
 
-    @log_calls('PluginManager')
     def _create_plugin_entry_in_global_config(self, plugin):
         app.config.add_per('plugins', plugin.short_name)
 
     def _remove_plugin_entry_in_global_config(self, plugin):
         app.config.del_per('plugins', plugin.short_name)
 
-    @log_calls('PluginManager')
+    def update_plugins(self, replace=True, activate=False, plugin_name=None):
+        '''
+        Move plugins from the downloaded folder to the user plugin folder
+
+        :param replace: replace plugin files if they already exist.
+        :type replace: boolean
+        :param activate: load and activate the plugin
+        :type activate: boolean
+        :param plugin_name: if provided, update only this plugin
+        :type plugin_name: str
+        :return: list of updated plugins (files have been installed)
+        :rtype: [] of str
+        '''
+        updated_plugins = []
+        user_dir = configpaths.get('PLUGINS_USER')
+        dl_dir = configpaths.get('PLUGINS_DOWNLOAD')
+        to_update = [plugin_name] if plugin_name else next(os.walk(dl_dir))[1]
+        for directory in to_update:
+            src_dir = os.path.join(dl_dir, directory)
+            dst_dir = os.path.join(user_dir, directory)
+            try:
+                if os.path.exists(dst_dir):
+                    if not replace:
+                        continue
+                    self.delete_plugin_files(dst_dir)
+                move(src_dir, dst_dir)
+            except Exception:
+                log.exception('Upgrade of plugin %s failed. Impossible to move '
+                    'files from "%s" to "%s"', directory, src_dir, dst_dir)
+                continue
+            updated_plugins.append(directory)
+            if activate:
+                pc = self.scan_dir_for_plugins(dst_dir, scan_dirs=True,
+                    package=True)
+                if not pc:
+                    continue
+                self.add_plugin(pc[0])
+                plugin = self.plugins[-1]
+                self.activate_plugin(plugin)
+        return updated_plugins
+
+
     def init_plugins(self):
         self._activate_all_plugins_from_global_config()
 
-    @log_calls('PluginManager')
     def add_plugin(self, plugin_class):
         '''
         :todo: what about adding plug-ins that are already added? Module reload
         and adding class from reloaded module or ignoring adding plug-in?
         '''
-        plugin = plugin_class()
+        try:
+            plugin = plugin_class()
+        except Exception:
+            log.exception('Error while loading a plugin')
+            return
 
         if plugin not in self.plugins:
             if not self._plugin_has_entry_in_global_config(plugin):
@@ -144,25 +187,27 @@ class PluginManager(metaclass=Singleton):
 
             self.plugins.append(plugin)
             plugin.active = False
+            app.nec.push_incoming_event(
+                NetworkEvent('plugin-added', plugin=plugin))
         else:
-            log.info('Not loading plugin %s v%s from module %s (identified by'
-                ' short name: %s). Plugin already loaded.' % (plugin.name,
-                plugin.version, plugin.__module__, plugin.short_name))
+            log.info('Not loading plugin %s v%s from module %s '
+                     '(identified by short name: %s). Plugin already loaded.',
+                     plugin.name, plugin.version,
+                     plugin.__module__, plugin.short_name)
 
-    @log_calls('PluginManager')
     def remove_plugin(self, plugin):
         '''
-        removes the plugin from the plugin list and deletes all loaded modules 
+        removes the plugin from the plugin list and deletes all loaded modules
         from sys. This way we will have a fresh start when the plugin gets added
-        agin.
+        again.
         '''
         if plugin.active:
             self.deactivate_plugin(plugin)
-            
+
         self.plugins.remove(plugin)
-        
+
         # remove modules from cache
-        base_package = plugin.__module__.split('.')[0]        
+        base_package = plugin.__module__.split('.')[0]
         # get the subpackages/-modules of the base_package. Add a dot to the
         # name to avoid name problems (removing module_abc if base_package is
         # module_ab)
@@ -171,27 +216,24 @@ class PluginManager(metaclass=Singleton):
         # remove the base_package itself
         if base_package in sys.modules:
             modules_to_remove.append(base_package)
-            
+
         for module_to_remove in modules_to_remove:
-            del sys.modules[module_to_remove] 
-        
-    @log_calls('PluginManager')
+            del sys.modules[module_to_remove]
+
     def add_plugins(self, plugin_classes):
         for plugin_class in plugin_classes:
             self.add_plugin(plugin_class)
 
-    @log_calls('PluginManager')
     def get_active_plugin(self, plugin_name):
         for plugin in self.active_plugins:
             if plugin.short_name == plugin_name:
                 return plugin
         return None
 
-    @log_calls('PluginManager')
     def extension_point(self, gui_extpoint_name, *args):
         '''
         Invokes all handlers (from plugins) for a particular extension point, but
-        doesnt add it to collection for further processing.
+        doesn't add it to collection for further processing.
         For example if you pass a message for encryption via extension point to a
         plugin, its undesired that the call is stored and replayed on activating the
         plugin. For example after an update.
@@ -207,7 +249,6 @@ class PluginManager(metaclass=Singleton):
         self._execute_all_handlers_of_gui_extension_point(gui_extpoint_name,
             *args)
 
-    @log_calls('PluginManager')
     def gui_extension_point(self, gui_extpoint_name, *args):
         '''
         Invokes all handlers (from plugins) for particular GUI extension point
@@ -237,7 +278,6 @@ class PluginManager(metaclass=Singleton):
         self._execute_all_handlers_of_gui_extension_point(gui_extpoint_name,
             *args)
 
-    @log_calls('PluginManager')
     def remove_gui_extension_point(self, gui_extpoint_name, *args):
         '''
         Removes GUI extension point from collection held by `PluginManager`.
@@ -279,8 +319,14 @@ class PluginManager(metaclass=Singleton):
                     self.gui_extension_points[gui_extpoint_name].remove(
                         ext_point)
 
+        if gui_extpoint_name not in self.gui_extension_points_handlers:
+            return
+
+        for handlers in self.gui_extension_points_handlers[gui_extpoint_name]:
+            disconnect_handler = handlers[1]
+            if disconnect_handler is not None:
+                disconnect_handler(args[0])
 
-    @log_calls('PluginManager')
     def _add_gui_extension_point_call_to_list(self, gui_extpoint_name, *args):
         '''
         Adds GUI extension point call to list of calls.
@@ -303,10 +349,9 @@ class PluginManager(metaclass=Singleton):
         '''
         if ((gui_extpoint_name not in self.gui_extension_points)
         or (args not in self.gui_extension_points[gui_extpoint_name])):
-            self.gui_extension_points.setdefault(gui_extpoint_name,[]).append(
+            self.gui_extension_points.setdefault(gui_extpoint_name, []).append(
                 args)
 
-    @log_calls('PluginManager')
     def _execute_all_handlers_of_gui_extension_point(self, gui_extpoint_name,
     *args):
         if gui_extpoint_name in self.gui_extension_points_handlers:
@@ -314,9 +359,9 @@ class PluginManager(metaclass=Singleton):
             gui_extpoint_name]:
                 try:
                     handlers[0](*args)
-                except Exception as e:
-                    log.warning('Error executing %s', handlers[0],
-                        exc_info=True)
+                except Exception:
+                    log.warning('Error executing %s',
+                                handlers[0], exc_info=True)
 
     def _register_events_handlers_in_ged(self, plugin):
         for event_name, handler in plugin.events_handlers.items():
@@ -351,7 +396,30 @@ class PluginManager(metaclass=Singleton):
         if plugin.encryption_name:
             del self.encryption_plugins[plugin.encryption_name]
 
-    @log_calls('PluginManager')
+    def _register_modules_with_handlers(self, plugin):
+        if not hasattr(plugin, 'modules'):
+            return
+        for con in app.connections.values():
+            for module in plugin.modules:
+                if not module.zeroconf and con.name == 'Local':
+                    continue
+                instance, name = module.get_instance(con)
+                modules.register_single_module(con, instance, name)
+
+                for handler in instance.handlers:
+                    con.connection.register_handler(handler)
+
+    def _unregister_modules_with_handlers(self, plugin):
+        if not hasattr(plugin, 'modules'):
+            return
+        for con in app.connections.values():
+            for module in plugin.modules:
+                instance = con.get_module(module.name)
+                modules.unregister_single_module(con, module.name)
+
+                for handler in instance.handlers:
+                    con.connection.unregister_handler(handler)
+
     def activate_plugin(self, plugin):
         '''
         :param plugin: plugin to be activated
@@ -364,6 +432,7 @@ class PluginManager(metaclass=Singleton):
             self._handle_all_gui_extension_points_with_plugin(plugin)
             self._register_events_handlers_in_ged(plugin)
             self._register_network_events_in_nec(plugin)
+            self._register_modules_with_handlers(plugin)
 
             self.active_plugins.append(plugin)
             try:
@@ -394,13 +463,14 @@ class PluginManager(metaclass=Singleton):
                     if handler:
                         try:
                             handler(*gui_extension_point_args)
-                        except Exception as e:
-                            log.warning('Error executing %s', handler,
-                                exc_info=True)
+                        except Exception:
+                            log.warning('Error executing %s',
+                                        handler, exc_info=True)
 
         self._remove_events_handler_from_ged(plugin)
         self._remove_network_events_from_nec(plugin)
         self._remove_name_from_encryption_plugins(plugin)
+        self._unregister_modules_with_handlers(plugin)
 
         # removing plug-in from active plug-ins list
         plugin.deactivate()
@@ -412,7 +482,6 @@ class PluginManager(metaclass=Singleton):
         for plugin_object in self.active_plugins:
             self.deactivate_plugin(plugin_object)
 
-    @log_calls('PluginManager')
     def _add_gui_extension_points_handlers_from_plugin(self, plugin):
         for gui_extpoint_name, gui_extpoint_handlers in \
         plugin.gui_extension_points.items():
@@ -423,7 +492,6 @@ class PluginManager(metaclass=Singleton):
         if plugin.encryption_name:
             self.encryption_plugins[plugin.encryption_name] = plugin
 
-    @log_calls('PluginManager')
     def _handle_all_gui_extension_points_with_plugin(self, plugin):
         for gui_extpoint_name, gui_extpoint_handlers in \
         plugin.gui_extension_points.items():
@@ -434,12 +502,11 @@ class PluginManager(metaclass=Singleton):
                     if handler:
                         try:
                             handler(*gui_extension_point_args)
-                        except Exception as e:
-                            log.warning('Error executing %s', handler,
-                                exc_info=True)
+                        except Exception:
+                            log.warning('Error executing %s',
+                                        handler, exc_info=True)
 
 
-    @log_calls('PluginManager')
     def _activate_all_plugins(self):
         '''
         Activates all plugins in `plugins`.
@@ -461,15 +528,31 @@ class PluginManager(metaclass=Singleton):
                 except GajimPluginActivateException:
                     pass
 
+    def register_modules_for_account(self, con):
+        '''
+        A new account has been added, register modules
+        of all active plugins
+        '''
+        for plugin in self.plugins:
+            if not plugin.active:
+                continue
+
+            if not hasattr(plugin, 'modules'):
+                continue
+
+            for module in plugin.modules:
+                instance, name = module.get_instance(con)
+                if not module.zeroconf and con.name == 'Local':
+                    continue
+                modules.register_single_module(con, instance, name)
+
     def _plugin_is_active_in_global_config(self, plugin):
         return app.config.get_per('plugins', plugin.short_name, 'active')
 
     def _set_plugin_active_in_global_config(self, plugin, active=True):
         app.config.set_per('plugins', plugin.short_name, 'active', active)
 
-    @staticmethod
-    @log_calls('PluginManager')
-    def scan_dir_for_plugins(path, scan_dirs=True, package=False):
+    def scan_dir_for_plugins(self, path, scan_dirs=True, package=False):
         r'''
         Scans given directory for plugin classes.
 
@@ -486,11 +569,10 @@ class PluginManager(metaclass=Singleton):
         :rtype: [] of class objects
 
         :note: currently it only searches for plugin classes in '\*.py' files
-                present in given direcotory `path` (no recursion here)
+                present in given directory `path` (no recursion here)
 
         :todo: add scanning zipped modules
         '''
-        from gajim.plugins.plugins_i18n import _
         plugins_found = []
         conf = configparser.ConfigParser()
         fields = ('name', 'short_name', 'version', 'description', 'authors',
@@ -528,44 +610,54 @@ class PluginManager(metaclass=Singleton):
                 try:
                     conf.read_file(conf_file)
                 except configparser.Error:
-                    log.warning(("Plugin {plugin} not loaded, error loading"
-                                 " manifest").format(plugin=elem_name)
-                                , exc_info=True)
+                    log.warning('Plugin %s not loaded, error loading manifest',
+                                elem_name, exc_info=True)
                     continue
 
+            short_name = conf.get('info', 'short_name', fallback=None)
+            if short_name is None:
+                log.error('No short_name defined for %s', elem_name)
+
+            # Check if the plugin is already loaded
+            try:
+                for plugin in self.plugins:
+                    if plugin.short_name == short_name:
+                        raise PluginAlreadyLoaded(
+                            'Skip Plugin %s because its '
+                            'already loaded' % elem_name)
+            except PluginAlreadyLoaded as error:
+                log.warning(error)
+                continue
+
             min_v = conf.get('info', 'min_gajim_version', fallback=None)
             max_v = conf.get('info', 'max_gajim_version', fallback=None)
 
-            gajim_v = app.config.get('version').split('+', 1)[0]
-            gajim_v_cmp = parse_version(gajim_v)
-
-            if min_v and gajim_v_cmp < parse_version(min_v):
-                log.warning(('Plugin {plugin} not loaded, newer version of'
-                             'gajim required: {gajim_v} < {min_v}').format(
-                                 plugin=elem_name,
-                                 gajim_v=gajim_v,
-                                 min_v=min_v
-                           ))
+            if min_v is None or max_v is None:
+                log.warning('Plugin without min/max version: %s', elem_name)
                 continue
-            if max_v and gajim_v_cmp > parse_version(max_v):
-                log.warning(('Plugin {plugin} not loaded, plugin incompatible '
-                             'with current version of gajim: '
-                             '{gajim_v} > {max_v}').format(
-                                 plugin=elem_name,
-                                 gajim_v=gajim_v,
-                                 max_v=max_v
-                           ))
+
+            gajim_v = gajim.__version__.split('+', 1)[0]
+            gajim_v_cmp = V(gajim_v)
+
+            if min_v and gajim_v_cmp < V(min_v):
+                log.warning('Plugin %s not loaded, newer version of'
+                            'gajim required: %s < %s',
+                            elem_name, gajim_v, min_v)
+                continue
+            if max_v and gajim_v_cmp > V(max_v):
+                log.warning('Plugin %s not loaded, plugin incompatible '
+                            'with current version of gajim: '
+                            '%s > %s', elem_name, gajim_v, max_v)
                 continue
 
             module = None
             try:
                 log.info('Loading %s', module_name)
                 module = __import__(module_name)
-            except Exception as error:
+            except Exception:
                 log.warning(
-                    "While trying to load {plugin}, exception occurred".format(plugin=elem_name),
-                    exc_info=sys.exc_info()
-                )
+                    'While trying to load %s, exception occurred',
+                    elem_name, exc_info=True)
                 continue
 
             if module is None:
@@ -575,7 +667,7 @@ class PluginManager(metaclass=Singleton):
             for module_attr_name in [attr_name for attr_name in dir(module)
             if not (attr_name.startswith('__') or attr_name.endswith('__'))]:
                 module_attr = getattr(module, module_attr_name)
-                log.debug('%s : %s' % (module_attr_name, module_attr))
+                log.debug('%s: %s', module_attr_name, module_attr)
 
                 try:
                     if not issubclass(module_attr, GajimPlugin) or \
@@ -586,37 +678,35 @@ class PluginManager(metaclass=Singleton):
                         os.path.dirname(file_path))
 
                     for option in fields:
-                        if conf.get('info', option) is '':
+                        if conf.get('info', option) == '':
                             raise configparser.NoOptionError(option, 'info')
                         if option == 'description':
-                            setattr(module_attr, option, _(conf.get('info', option)))
+                            setattr(module_attr, option, plugins_i18n._(conf.get('info', option)))
                             continue
                         setattr(module_attr, option, conf.get('info', option))
 
                     plugins_found.append(module_attr)
                 except TypeError:
-                    # set plugin localization
-                    try:
-                        module_attr._ = _
-                    except AttributeError:
-                        pass
+                    pass
                 except configparser.NoOptionError:
                     # all fields are required
-                    log.debug('%s : %s' % (module_attr_name,
-                        'wrong manifest file. all fields are required!'))
+                    log.debug(
+                        '%s: wrong manifest file. all fields are required!',
+                        module_attr_name)
                 except configparser.NoSectionError:
                     # info section are required
-                    log.debug('%s : %s' % (module_attr_name,
-                        'wrong manifest file. info section are required!'))
+                    log.debug(
+                        '%s: wrong manifest file. info section are required!',
+                        module_attr_name)
                 except configparser.MissingSectionHeaderError:
                     # info section are required
-                    log.debug('%s : %s' % (module_attr_name,
-                        'wrong manifest file. section are required!'))
+                    log.debug('%s: wrong manifest file. section are required!',
+                              module_attr_name)
 
         sys.path.remove(path)
         return plugins_found
 
-    def install_from_zip(self, zip_filename, owerwrite=None):
+    def install_from_zip(self, zip_filename, overwrite=None):
         '''
         Install plugin from zip and return plugin
         '''
@@ -653,7 +743,7 @@ class PluginManager(metaclass=Singleton):
 
         if os.path.isdir(plugin_dir):
         # Plugin dir already exists
-            if not owerwrite:
+            if not overwrite:
                 raise PluginsystemError(_('Plugin already exists'))
             self.uninstall_plugin(self.get_plugin_by_path(plugin_dir))
 
@@ -667,10 +757,7 @@ class PluginManager(metaclass=Singleton):
         plugin = self.plugins[-1]
         return plugin
 
-    def uninstall_plugin(self, plugin):
-        '''
-        Deactivate and remove plugin from `plugins` list
-        '''
+    def delete_plugin_files(self, plugin_path):
         def on_error(func, path, error):
             if func == os.path.islink:
             # if symlink
@@ -679,13 +766,28 @@ class PluginManager(metaclass=Singleton):
             # access is denied or other
             raise PluginsystemError(error[1][1])
 
-        if plugin:
-            self.remove_plugin(plugin)
-            rmtree(plugin.__path__, False, on_error)
-            if self._plugin_has_entry_in_global_config(plugin):
-                self._remove_plugin_entry_in_global_config(plugin)
+        rmtree(plugin_path, False, on_error)
+
+    def uninstall_plugin(self, plugin):
+        '''
+        Deactivate and remove plugin from `plugins` list
+        '''
+        if not plugin:
+            return
+
+        self.remove_plugin(plugin)
+        self.delete_plugin_files(plugin.__path__)
+        if self._plugin_has_entry_in_global_config(plugin):
+            self._remove_plugin_entry_in_global_config(plugin)
+
+        app.nec.push_incoming_event(
+            NetworkEvent('plugin-removed', plugin=plugin))
 
     def get_plugin_by_path(self, plugin_dir):
         for plugin in self.plugins:
             if plugin.__path__ in plugin_dir:
                 return plugin
+
+
+class PluginAlreadyLoaded(Exception):
+    pass
